;;;; "opticompute.scm" Optical stack and parameter extraction. -*-scheme-*-
;;; Copyright (C) 2003, 2004 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

;; http://localhost/~jaffer/FreeSnell

(require 'array)
(require-if 'compiling 'printf)		; for DESCRIBE-STACK

(define (avg2 x y) (/ (+ x y) 2))

(define (parameter-name tok)
  (case tok
    ((T_s) 's-transmission)
    ((R_s) 's-reflection)
    ((B_s) 's-reflection-back)
    ((T_p) 'p-transmission)
    ((R_p) 'p-reflection)
    ((B_p) 'p-reflection-back)
    ((T)   'tranmission)
    ((R)   'reflection)
    ((B)   'reflection-back)
    ((A)   'absorption)
    ((L)   'absorption-back)
    (else (slib:error 'parameter-name 'unknown 'token tok))))

(define (extract-parameter tok db? spol ppol)
  ((if db? dbize identity)
   (case tok
     ((1) 1)
     ((T_s) (car spol))
     ((R_s) (cadr spol))
     ((B_s) (caddr spol))
     ((T_p) (car ppol))
     ((R_p) (cadr ppol))
     ((B_p) (caddr ppol))
     ((T)   (avg2 (car spol) (car ppol)))
     ((R)   (avg2 (cadr spol) (cadr ppol)))
     ((B)   (avg2 (caddr spol) (caddr ppol)))
     ((A)   (- 1 (avg2 (car spol) (car ppol))
	       (avg2 (cadr spol) (cadr ppol))))
     ((L)   (- 1 (avg2 (car spol) (car ppol))
	       (avg2 (caddr spol) (caddr ppol))))
     (else (slib:error 'extract-parameter 'unknown 'token tok)))))

(define (extract-refractive w tok db? z)
  ((if db? dbize identity)
   (case tok
     ((n real real-part) (real-part z))
     ((k imag imag-part) (imag-part z))
     ((ec extinction-coefficient) (/ (imag-part z) (real-part z)))
     ((r0 reflect reflectance)
      (let ((spol (combine-layers 0 (- w) 1 '() z))
	    (ppol (combine-layers 0 w 1 '() z)))
	(avg2 (cadr spol) (cadr ppol))))
     (else (slib:warn 'extract-refractive 'unknown 'token tok)))))

(define (compute/wavelength wmin wmax w-transform samples db? incidents IRs
			    n1 layers n2)
  (define cpi (count-plots incidents))
  (define data (create-array (Ar64)
			     samples
			     (+ 1 (count-plots incidents) (count-plots IRs))))
  (define (average/incident angles w)
    (define cang (length angles))
    (define vals (combine-layers (car angles) w n1 layers n2))
    (cond ((eqv? 1 cang) vals)
	  (else
	   (for-each
	    (lambda (angle)
	      (set! vals (map + vals (combine-layers angle w n1 layers n2))))
	    (cdr angles))
	   (map (lambda (x) (/ x cang)) vals))))
  (define (iterate sdx w)
    (define idx 0)
    (array-set! data (w-transform w) sdx 0)
    (for-each
     (lambda (incident)
       (let ((ang (car incident)))
	 (define spol (average/incident ang (- w)))
	 (define ppol (average/incident ang w))
	 (for-each
	  (lambda (tok)
	    (set! idx (+ 1 idx))
	    (array-set! data (extract-parameter tok db? spol ppol) sdx idx))
	  (cdr incident))))
     incidents)
    (for-each
     (lambda (IR)
       (define z (if (number? (car IR)) (car IR) ((car IR) w)))
       (for-each
	(lambda (tok)
	  (set! idx (+ 1 idx))
	  (array-set! data (extract-refractive w tok db? z) sdx idx))
	(cdr IR)))
     IRs))
  (if (< 1.5 (/ wmax wmin) 10000)
      (let ((winc (expt (/ wmax wmin) (/ (+ -1 samples)))))
	(do ((cnt (+ -1 samples) (+ -1 cnt))
	     (w wmin (* winc w)))
	    ((negative? cnt))
	  (iterate cnt w)))
      (let ((winc (/ (- wmax wmin) (+ -1 samples))))
	(do ((cnt (+ -1 samples) (+ -1 cnt))
	     (w wmin (+ winc w)))
	    ((negative? cnt))
	  (iterate cnt w))))
  data)

(define (compute/angle thmin thmax rad->deg samples db? wvspecs n1 layers n2)
  (define data (create-array (Ar64) samples (+ 1 (count-plots wvspecs))))
  (define thinc (/ (- thmax thmin) (+ -1 samples)))
  (do ((cnt (+ -1 samples) (+ -1 cnt))
       (th thmin (+ thinc th)))
      ((negative? cnt))
    (array-set! data (rad->deg th) cnt 0)
    (let ((idx 0))
      (for-each
       (lambda (wvspec)
	 (define spol (combine-layers th (- (car wvspec)) n1 layers n2))
	 (define ppol (combine-layers th (car wvspec) n1 layers n2))
	 (for-each
	  (lambda (tok)
	    (set! idx (+ 1 idx))
	    (array-set! data (extract-parameter tok db? spol ppol) cnt idx))
	  (cdr wvspec)))
       wvspecs)))
  data)

(define (optical-stack . args)
  (define layers '())
  (define wnom #f)
  (define n1 #f)
  (define n2 #f)
  (define (do-arg arg)
    (cond ((not (and (list? arg) (pair? arg)))
	   (slib:warn 'optical-stack 'mystery arg))
	  (else (case (car arg)
		  ((repeat)
		   (do ((cnt (+ -1 (cadr arg)) (+ -1 cnt)))
		       ((negative? cnt))
		     (for-each do-arg (cddr arg))))
		  ((nominal)
		   (set! wnom (cadr arg)))
		  ((layer)
		   (let ((thick (caddr arg)))
		     (if (not n1) (set! n1 1))
		     (if (not (null? (cdddr arg)))
			 (set! wnom (cadddr arg)))
		     (set! layers
			   (cons
			    (if (> thick 1/100)
				(list (cadr arg)
				      (/ (* wnom thick)
					 (real-part
					  (if (number? (cadr arg))
					      (cadr arg)
					      ((cadr arg) wnom))))
				      wnom)
				(if wnom
				    (list (cadr arg) thick wnom)
				    (cdr arg)))
			    layers))))
		  ((substrate)
		   (if n2 (slib:error 'too-many 'substrate arg))
		   (if n1
		       (set! n2 (cadr arg))
		       (set! n1 (cadr arg))))
		  (else
		   (slib:warn 'optical-stack 'unknown arg))))))
  (for-each do-arg args)
  (if (not n2) (set! n2 1))
  (if (not n1) (set! n1 n2))
  (list n1 (merge-layers layers) n2))

;;;also reverses
(define (merge-layers layers)
  (define (loop layers lay lays)
    (cond ((null? layers) (cons lay lays))
	  ((and (eqv? (caar layers) (car lay))
		(eqv? (caddar layers) (caddr lay)))
	   (loop (cdr layers)
		 (list (car lay)
		       (+ (cadar layers) (cadr lay))
		       (caddr lay))
		 lays))
	  (else (loop (cdr layers) (car layers) (cons lay lays)))))
  (if (null? layers)
      layers
      (loop (cdr layers) (car layers) '())))

(define (describe-stack n1 layers n2)
  (define (frac-wave frac)
    (define inv (/ 1 frac))
    (if (and (real? frac)
	     (< 1/100 frac 1)
	     (< (abs (- inv (inexact->exact (round inv)))) (* .01 inv)))
	(sprintf #f "1/%d" (inexact->exact inv))
	(sprintf #f "%.3f" frac)))
  (require 'printf)
  (cond ((number? n1) (printf "Top substrate IR = %g\\n" n1)))
  (for-each
   (lambda (layer)
     (cond ((and (number? (car layer)) (not (null? (cddr layer))))
	    (printf "IR = %-11.3g %6.3Km thick (%6.3Km optical) = %5s wave @ %6.3Km\\n"
		    (car layer)
		    (cadr layer)
		    (* (cadr layer) (real-part (car layer)))
		    (frac-wave (/ (* (cadr layer) (real-part (car layer)))
				  (caddr layer)))
		    (caddr layer)))
	   ((number? (car layer))
	    (printf "IR = %-11.3g %6.3Km thick (%6.3Km optical)\\n"
		    (car layer)
		    (cadr layer)
		    (* (cadr layer) (real-part (car layer)))))
	   ((not (null? (cddr layer)))
	    (let ((rough ((car layer) (caddr layer))))
	      (printf "IR ~ %-11.3g %6.3Km thick (%6.3Km optical) ~ %5s wave @ %6.3Km\\n"
		      rough
		      (cadr layer)
		      (* (cadr layer) (real-part rough))
		      (frac-wave (/ (* (cadr layer) (real-part rough))
				    (caddr layer)))
		      (caddr layer))))
	   (else
	    (printf "IR ~ %-11s %6.3Km thick\\n" "?" (cadr layer)))))
   layers)
  (cond ((number? n2) (printf "Bottom substrate IR = %g\\n" n2))))

(define (substrate index-of-refraction)
  (list 'substrate index-of-refraction))

(define (layer index-of-refraction thickness . nominal)
  (apply list 'layer index-of-refraction thickness nominal))

;;(define (metallic index-of-refraction thickness)
;;  (list 'layer index-of-refraction thickness))

(define (nominal wl)
  (list 'nominal wl))

(define (repeat . args)
  (cons 'repeat args))
;;(trace-all "../FreeSnell/opticompute.scm") (set! *qp-width* 333)
