;;;; "optiplot.scm" graphs and drives the calculations in "opticompute.scm" -*-scheme-*-
;;; Copyright (C) 2003, 2004 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

;; http://localhost/~jaffer/FreeSnell

(require 'fresnel-equations)
(require 'array)
(require 'subarray)			; for SMOOTH-COLUMNS and COLOR
(require 'array-for-each)		; for SMOOTH-COLUMNS
(require 'color-space)
(require 'eps-graph)
(require 'printf)			; for DESCRIBE-PLT
(require 'sort)
(require 'pnm)				; for color patches

;;; Defaults from graph parameters
(define graph:dimensions '(600 400))
(define graph:font "Helvetica")
(define graph:font-size 12)

(define (ps:viewer path)
  (system (string-append "gv '" path "'")))

;;; View using ps:viewer or convert using ImageMagick
(define *output-format* #t)
(define (convert-output path suffix)
  (define (replace.suf path suf)
    (string-append (substring path 0 (+ -3 (string-length path)))
		   (if (symbol? suf) (symbol->string suf) suf)))
  (case suffix
    ((#t) (and ps:viewer (ps:viewer path)) path)
    ((eps) path)
    ((ps)
     (let ((psname (replace.suf path "ps")))
       (if (file-exists? psname) (delete-file psname))
       (rename-file path psname)
       psname))
    ((pdf)
     (and (zero? (system (string-append "epstopdf '" path "'")))
	  (delete-file path)
	  (replace.suf path "pdf")))
    ((png jpg jpeg)
     (let ((sufname (replace.suf path suffix)))
       (and (zero? (system (string-append "convert '" path "' '" sufname "'")))
	    (delete-file path)
	    sufname)))
    ((#f) path)
    (else (slib:error 'convert-output 'unknown 'suffix suffix))))

;;; Angle conversions
(define /180pi (/ 180 pi))
(define /pi180 (/ pi 180))
(define (rad->deg rad) (* /180pi rad))
(define (deg->rad rad) (* /pi180 rad))
				 ; Energy	     E =         h*c/L
(define eV<->L			 ; Photon wavelength L =         h*c/E
  (let ((c 299792458)		 ; speed of light    c =           m/s
	(h 6.62606876e-34)	 ; Plank's constant  h =           J.s
	(J/eV 1.602176462e-19))
    (lambda (eV) (/ (* h c) eV J/eV))))

(define log_10
  (let ((log10 (log 10)))
    (lambda (x) (/ (log x) log10))))

(define (dbize x)
  (if (or (not (real? x)) (negative? x)) (slib:error 'dbize x))
  (* 20 (log_10 x)))

(define (parameter-color tok)
  (case tok
    ((T_s) 'gigas)
    ((R_s) 'golddrop)
    ((B_s) 'goldtips)
    ((T_p) 'gagas)
    ((R_p) 'golddrop)
    ((B_p) 'goldtips)
    ((T)   'gigas)
    ((R)   'golddrop)
    ((B)   'butteredrum)
    ((A)   'goblin)
    ((L)   'goldensand)
    (else (slib:warn 'parameter-color 'unknown 'token tok) 'black)))

(define (count-plots incidents)
  (apply + (map length (map cdr incidents))))

(define (smooth-columns data)
  (define dims (array-dimensions data))
  (let ((rng (list 1 (+ -1 (cadr dims)))))
    (define nra (apply create-array data (array-dimensions data)))
    (do ((idx (+ -1 (car dims)) (+ -1 idx))
	 (rowi-1 (subarray data (+ -2 (car dims)) rng)
		 (and (not (negative? (+ -2 idx)))
		      (subarray data (+ -2 idx) rng)))
	 (rowi (subarray data (+ -1 (car dims)) rng) rowi-1)
	 (rowi+1 #f rowi)
	 (nrwi (subarray nra (+ -1 (car dims)) rng)
	       (and (not (negative? (+ -1 idx)))
		    (subarray nra (+ -1 idx) rng))))
	((negative? idx) nra)
      (array-set! nra (array-ref data idx 0) idx 0)
      (cond ((or (not rowi+1) (not rowi-1)) (array-map! nrwi identity rowi))
	    (else (array-map! nrwi + rowi-1 rowi rowi rowi+1)
		  (array-map! nrwi * 1/4 nrwi))))))

;;;Most recently computed spectra
(define *trp-spectra* #f)

(define (plot-response . args)
  (define db? #f)
  (define rmin #f)
  (define rmax #f)
  (define x-scale-type #f)
  (define incidents '())
  (define wavelengths '())
  (define IRs '())
  (define markers '())
  (define samples #f)
  (define title #f)
  (define file #f)
  (define wmin #f)
  (define wmax #f)
  (define thmin #f)
  (define thmax #f)
  (define optics #f)
  (define smooth? #f)
  (define colors '())
  (define outtype *output-format*)
  (fluid-let ((graph:dimensions graph:dimensions)
	      (graph:font-size graph:font-size)
	      (graph:font graph:font))
    (for-each
     (lambda (arg)
       (cond ((not (and (list? arg) (pair? arg)))
	      (slib:warn 'plot-response 'mystery arg))
	     ((number? (car arg))
	      (set! optics arg))
	     (else (case (car arg)
		     ((title)
		      (set! title (cadr arg))
		      (set! file (caddr arg)))
		     ((range)
		      (set! rmin (cadr arg))
		      (set! rmax (caddr arg)))
		     ((db)
		      (set! db? #t)
		      (cond ((null? (cdr arg)))
			    (else (set! rmin (cadr arg))
				  (set! rmax (caddr arg)))))
		     ((smooth)
		      (set! smooth? #t))
		     ((samples)
		      (set! samples (cadr arg)))
		     ((color)
		      (set! colors (cons (cdr arg) colors)))
		     ((linear)
		      (set! x-scale-type 'linear)
		      (set! wmin (cadr arg))
		      (set! wmax (caddr arg)))
		     ((angles)
		      (set! x-scale-type 'angle)
		      (set! thmin (cadr arg))
		      (set! thmax (caddr arg)))
		     ((logarithmic)
		      (set! x-scale-type 'logarithmic)
		      (set! wmin (cadr arg))
		      (set! wmax (caddr arg)))
		     ((wavenumber)
		      (set! x-scale-type 'wavenumber)
		      (set! wmin (cadr arg))
		      (set! wmax (caddr arg)))
		     ((eV)
		      (set! x-scale-type 'eV)
		      (set! wmin (eV<->L (caddr arg)))
		      (set! wmax (eV<->L (cadr arg))))
		     ((marker)
		      (set! markers (append (cdr arg) markers)))
		     ((incident)
		      (set! incidents (cons (cdr arg) incidents)))
		     ((wavelength)
		      (set! wavelengths (cons (cdr arg) wavelengths)))
		     ((IR)
		      (set! IRs (cons (cdr arg) IRs)))
		     ((output-format)
		      (let ((arg-2
			     (cond ((symbol? (cadr arg))
				    (set! outtype (cadr arg))
				    (cddr arg))
				   (else (cdr arg)))))
			(case (length arg-2)
			  ((0) #t)
			  ((1)
			   (cond ((symbol? (car arg-2))
				  (set! graph:font (car arg-2)))
				 (else
				  (set! graph:dimensions (car arg-2))
				  (set! graph:font-size (/ (car arg-2) 33)))))
			  ((2)
			   (cond ((symbol? (car arg-2))
				  (set! graph:font (car arg-2))
				  (set! graph:font-size (cadr arg-2)))
				 (else
				  (set! graph:dimensions arg-2)
				  (set! graph:font-size (/ (car arg-2) 33)))))
			  ((3)
			   (set! graph:dimensions
				 (list (car arg-2) (cadr arg-2)))
			   (cond ((symbol? (caddr arg-2))
				  (set! graph:font (caddr arg-2))
				  (set! graph:font-size (/ (car arg-2) 33)))
				 (else
				  (set! graph:font-size (caddr arg-2)))))
			  ((4)
			   (set! graph:dimensions
				 (list (car arg-2) (cadr arg-2)))
			   (set! graph:font (caddr arg-2))
			   (set! graph:font-size (cadddr arg-2))))))
		     (else
		      (slib:warn 'plot-response 'unknown arg))))))
     args)
    (set! wavelengths (reverse wavelengths))
    (set! incidents (reverse incidents))
    (set! colors (reverse colors))
    (set! IRs (reverse IRs))
    (if (not samples) (set! samples 200))
    ;;(write file) (display ": ") (display title)
    (newline)
    (describe-stack (car optics) (cadr optics) (caddr optics))
    (if (not (null? colors))
	(make-colors-swatch file colors optics))
    (let ((path (string-append file ".eps"))) ; "/tmp/"
      (define w-transform
	(case x-scale-type
	  ((linear) identity)
	  ((logarithmic) log_10)
	  ((wavenumber) (lambda (w) (log_10 (/ .01 w))))
	  ((eV) eV<->L)
	  ((angle) identity)
	  ;;(else (slib:error 'plot-response 'unknown 'x-scale-type x-scale-type))
	  ))
      (cond
       ((not x-scale-type))
       ((null? wavelengths)
	(describe-plt wmin wmax x-scale-type samples markers w-transform
		      incidents IRs path title)
	(case x-scale-type
	  ((linear logarithmic)
	   (set! *trp-spectra*
		 (apply compute/wavelength wmin wmax w-transform
			samples db? incidents IRs optics))
	   (if smooth? (set! *trp-spectra* (smooth-columns *trp-spectra*)))
	   (plot/wavelength *trp-spectra*
			    (w-transform wmin) (w-transform wmax) x-scale-type
			    rmin rmax db? incidents IRs
			    (map w-transform markers) path title))
	  ((wavenumber eV)
	   (set! *trp-spectra*
		 (apply compute/wavelength wmax wmin w-transform
			samples db? incidents IRs optics))
	   (plot/wavelength *trp-spectra*
			    (w-transform wmax) (w-transform wmin) x-scale-type
			    rmin rmax db? incidents IRs
			    (map w-transform markers) path title)
	   )))
       (else (plot/angle
	      (apply compute/angle thmin thmax rad->deg
		     samples db? wavelengths optics)
	      (rad->deg thmin) (rad->deg thmax)
	      rmin rmax db? wavelengths path title)))

      (and x-scale-type (convert-output path outtype)))))

(define (x-scale-ruler x-scale-type)
  (case x-scale-type
    ((logarithmic)
     (rule-horizontal bottomedge "log(wavelength/m)" 10))
    ((linear)
     (rule-horizontal bottomedge "wavelength in m" 10))
    ((wavenumber)
     (rule-horizontal bottomedge "log(wavenumber*cm)" 10))
    ((eV)
     (rule-horizontal bottomedge "energy in eV" 10))
    ((angle)
     (rule-horizontal bottomedge "angle in degrees" 10))
    (else (slib:error 'x-scale-ruler 'unknown x-scale-type))))

(define (find-encompassing-range Data start end)
  (do ((idx (+ -1 end) (+ -1 idx))
       (lst '() (cons (column-range Data idx) lst)))
      ((< idx start) (apply combine-ranges lst))))

(define (plot-columns Data idx incidents colorizer glyphizer)
  (set! idx (+ -1 idx))
  (map
   (lambda (incident)
     (map
      (lambda (tok)
	(set! idx (+ 1 idx))
	(in-graphic-context
	 (set-color (colorizer tok))
	 (plot-column Data 0 idx (glyphizer tok))))
      (cdr incident)))
   incidents))

(define (plot/angle Data thmin thmax rmin rmax db? wavelengths path title)
  (define cpi (count-plots wavelengths))
  (create-postscript-graph
   path graph:dimensions
   (set-font graph:font graph:font-size)
   (whole-page)
   (setup-plot
    (list thmin thmax)
    (if rmin
	(list rmin rmax)
	(snap-range (find-encompassing-range Data 1 (+ 1 cpi)))))
   (title-top "" title)
   (outline-rect plotrect)
   (in-graphic-context (set-linewidth 0) (grid-verticals) (grid-horizontals))
   (x-scale-ruler 'angle)
   (rule-vertical leftedge (if db? "dB" "power ratio") 10)
   (plot-columns Data 1 wavelengths
		 parameter-color (lambda (tok) 'line))))

(define (compute/markers markers)
  (if (null? markers)
      #f
      (let ((Data (create-array (Ar64 50) (length markers) 2))
	    (idx 0))
	(for-each (lambda (mark)
		    (array-set! Data mark idx 0)
		    (set! idx (+ 1 idx)))
		  markers)
	Data)))

(define (plot/markers Data wmin wmax)
  (if Data
      (in-graphic-context
       (setup-plot (list wmin wmax) '(0 100) 'graphrect)
       (plot-column Data 0 1 'impulse)
       (plot-column Data 0 1 'triup))
      ""))

(define (plot/wavelength Data wmin wmax x-scale-type
			 rmin rmax db? incidents IRs markers file title)
  (define (collect-edges edg)
    (if db?
	(string-append "20*log(" (symbol->string edg) ")")
	(symbol->string edg)))
  (define (comma-append strs)
    (cond ((null? strs) "")
	  ((null? (cdr strs)) (car strs))
	  (else
	   (string-append (car strs) ", "
			  (comma-append (cdr strs))))))
  (define mdata (compute/markers markers))
  (define ledges '())
  (define redges '())
  (let ((cpi (count-plots incidents)))
    (define tra-plot
      (if (null? incidents)
	  ""
	  (in-graphic-context
	   (setup-plot
	    (list wmin wmax)
	    (if rmin
		(list rmin rmax)
		(snap-range (find-encompassing-range Data 1 (+ 1 cpi)))))
	   (title-top "" title)
	   (outline-rect plotrect)
	   (in-graphic-context (set-linewidth 0) (grid-verticals) (grid-horizontals))
	   (x-scale-ruler x-scale-type)
	   (rule-vertical leftedge (if db? "dB" "power ratio") 10)
	   (plot-columns Data 1 incidents
			 parameter-color (lambda (tok) 'line))
	   (plot/markers mdata wmin wmax))))
    (define ir-plot
      (if (null? IRs)
	  ""
	  (let ((columns
		 (plot-columns
		  Data (+ 1 cpi) IRs
		  (lambda (tok)
		    (case tok
		      ((n real real-part)
		       (if (not (memq 'n ledges))
			   (set! ledges (cons 'n ledges)))
		       'AntiqueBrass)
		      ((k imag imag-part)
		       (if (not (memq 'k redges))
			   (set! redges (cons 'k redges)))
		       'seagreen)
		      ((ec extinction-coefficient)
		       (if (not (memq 'k/n redges))
			   (set! redges (cons 'k/n redges)))
		       'seagreen)
		      ((r0 reflect reflectance)
		       (if (not (memq 'r0 ledges))
			   (set! ledges (cons 'r0 ledges)))
		       'golddrop)
		      (else (slib:warn 'plot/wavelength
				       'unknown 'token tok) 'black)))
		  (lambda (tok) 'line))))
	    (in-graphic-context
	     (setup-plot (list wmin wmax)
			 (snap-range (find-encompassing-range
				      Data (+ 1 cpi) (+ 1 cpi (count-plots IRs)))))
	     (title-top "" (if (null? incidents) title ""))
	     (outline-rect plotrect)
	     (in-graphic-context (set-linewidth 0) (grid-verticals) (grid-horizontals))
	     (x-scale-ruler x-scale-type)
	     (if (null? ledges)
		 ""
		 (in-graphic-context
		  (set-color 'AntiqueBrass)
		  (rule-vertical leftedge
				 (comma-append (map collect-edges ledges))
				 10)))
	     (if (null? redges)
		 ""
		 (in-graphic-context
		  (set-color 'seagreen)
		  (rule-vertical rightedge
				 (comma-append (map collect-edges redges))
				 -10)))
	     columns
	     (plot/markers mdata wmin wmax)))))
    (if (null? IRs)
	(create-postscript-graph file graph:dimensions
				 (set-font graph:font graph:font-size)
				 (whole-page) tra-plot)
	(create-postscript-graph
	 file (if (null? incidents)
		  graph:dimensions
		  (list (car graph:dimensions)
			(* 2 (cadr graph:dimensions))))
	 (set-font graph:font graph:font-size)
	 (whole-page)
	 (if (null? incidents) "" (partition-page 1 2))
	 tra-plot
	 ir-plot))))


(define (describe-plt wmin wmax x-scale-type samples markers w-transform
		      incidents IRs path title)
  (printf "\"%s\"\\n" title)
  (printf "Plotting %d samples from %6.3Km :: %6.3Km (%s)\\n"
	  samples wmin wmax x-scale-type)
  (cond ((not (null? markers))
	 (printf " with markers at")
	 (for-each (lambda (marker) (printf " %6.3Km" marker))
		   (sort markers <))
	 (newline)))
  (printf " to \"%s\"\\n" path))

(define (title str . file)
  (list 'title str (if (null? file) str (car file))))

(define (incident zdegrees . plot-list)
  (cons 'incident (cons (if (number? zdegrees)
			    (list (deg->rad zdegrees))
			    (map deg->rad zdegrees))
			plot-list)))

;;; Each angle th is weighted by 2*pi*sin(th)
;;; The average angle over the ZR cone is (sin(zr)-(zr*cos(zr)))/(1-cos(zr));
;;; very close to 0.65*zr
(define (cone zdegrees . plot-list)
  (define zr (deg->rad zdegrees))
  (cons 'incident (cons (list (/ (- (sin zr) (* zr (cos zr)))
				 (- 1 (cos zr))))
			plot-list)))

(define (wavelength wl . plot-list)
  (cons 'wavelength (cons wl plot-list)))

(define (IR substance . plot-list)
  (cons 'IR (cons substance plot-list)))

(define (samples count)
  (list 'samples count))

(define (marker . ws)
  (cons 'marker ws))

(define (angles ang . args)
  (define ang/4 (/ (deg->rad ang) 4))
  (set! ang (deg->rad ang))
  (set! args (map deg->rad args))
  (case (length args)
    ((0) (list 'angles (- ang ang/4) (+ ang ang/4)))
    ((1) (list 'angles ang (car args)))
    (else (slib:error 'angles ang args))))

(define (linear wavelength . args)
  (define wl/4 (/ wavelength 4))
  (case (length args)
    ((0) (list 'linear (- wavelength wl/4) (+ wavelength wl/4)))
    ((1 2) (cons 'linear (cons wavelength args)))
    (else (slib:error 'linear wavelength args))))

(define (logarithmic wavelength . args)
  (case (length args)
    ((0) (list 'logarithmic (/ wavelength 1.5) (* wavelength 1.5)))
    ((1 2) (cons 'logarithmic (cons wavelength args)))
    (else (slib:error 'logarithmic wavelength args))))

(define (color angle tok)
  (list 'color (deg->rad angle) tok))

(define (wavenumber x . args)
  (case (length args)
    ((0) (list 'wavenumber (/ x 1.5) (* x 1.5)))
    ((1 2) (cons 'wavenumber (cons x args)))
    (else (slib:error 'wavenumber x args))))

(define (eV x . args)
  (case (length args)
    ((0) (list 'eV (/ x 1.5) (* x 1.5)))
    ((1 2) (cons 'eV (cons x args)))
    (else (slib:error 'eV x args))))

(define (range low high . units)
  (list (if (null? units) 'range (car units)) low high))

(define (output-format . args)
  (cons 'output-format args))

(define (db . args)
  (cons 'db args))

(define (smooth)
  (list 'smooth))
;;(trace-all "../FreeSnell/optiplot.scm") (set! *qp-width* 333)
